#ifdef _WIN32
#include <winsock.h>
typedef int socklen_t;
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <memory.h>
#define INVALID_SOCKET (-1)
#define SOCKET_ERROR (-1)
typedef int SOCKET;
#endif
#include <cstdint>

bool send_cmd_packet(SOCKET s, uint32_t cmd, uint32_t attach_size = 0, uint32_t offset = 0, uint32_t size = 0, bool reply = false);
bool recv_data(SOCKET s, char* buf, size_t nbytes, bool accept_incomplete_packet = false, size_t* packet_size = nullptr);
bool send_data(SOCKET s, const char* buf, size_t nbytes);
bool udp_recv_datagram(SOCKET s, char* buf, size_t max_length, size_t* datagram_size, struct sockaddr_in* from_addr = nullptr);
bool set_sock_recvtimeo(SOCKET s, int msec);

bool send_cmd_packet(SOCKET s, uint32_t cmd, uint32_t attach_size, uint32_t offset, uint32_t size, bool reply)
{
    int nret;
    uint32_t blk[4];
    uint16_t rx;
    blk[0] = cmd; blk[1] = attach_size; blk[2] = offset; blk[3] = size;
    nret = send(s, reinterpret_cast<char*>(&blk), sizeof(blk), 0);
    if (nret != sizeof(blk)) {
        return false;
    }
    if (reply) {
        nret = recv(s, reinterpret_cast<char*>(&rx), sizeof(rx), 0);
        return (nret == sizeof(rx) && rx == 0x4B4F);
    }
    return true;
}

bool recv_data(SOCKET s, char *buf, size_t nbytes, bool accept_incomplete_packet, size_t *packet_size)
{
    int nret, nrecv = 0;
    if (!buf || !nbytes) {
        return false;
    }
    do {
        nret = recv(s, &buf[nrecv], nbytes-nrecv, 0);
        if (nret > 0) {
            nrecv += nret;
        }
    } while(nret != SOCKET_ERROR && !accept_incomplete_packet && (nrecv < nbytes));
    if (accept_incomplete_packet) {
        if (packet_size)
            *packet_size = nrecv;
        return (nrecv > 0);
    }
    else {
        return (nrecv == nbytes);
    }
}

bool send_data(SOCKET s, const char *buf, size_t nbytes)
{
    int nret;
    if (!buf || !nbytes) {
        return false;
    }
    nret = send(s, buf, nbytes, 0);
    return (nret == nbytes);
}

bool udp_recv_datagram(SOCKET s, char *buf, size_t max_length, size_t *datagram_size, sockaddr_in *from_addr)
{
    int nret;
    socklen_t from_size;
    struct sockaddr_in from_addr_l;
    from_size = sizeof(from_addr_l);
    if (s == INVALID_SOCKET) {
        return false;
    }
    nret = recvfrom(s, buf, max_length, 0, reinterpret_cast<struct sockaddr*>(&from_addr_l), &from_size);
    if (nret > 0)
    {
        if (from_addr) {
            memcpy(from_addr, &from_addr_l, sizeof(from_addr_l));
        }
        *datagram_size = nret;
        return true;
    }
    else
    {
        return false;
    }
}

bool set_sock_recvtimeo(SOCKET s, int msec)
{
#ifdef _WIN32
    DWORD t = msec;
#else
    timeval t;
    t.tv_sec = msec / 1000;
    t.tv_usec = (msec % 1000) * 1000;
#endif
    return (setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, reinterpret_cast<char*>(&t), sizeof(t)));
}

